libdfp {
  LIBDFP_1.0.0
  LIBDFP_1.0.2
  LIBDFP_1.0.3
  LIBDFP_1.0.4
  LIBDFP_1.0.5
  LIBDFP_1.0.6
  LIBDFP_1.0.7
  LIBDFP_1.0.8
  LIBDFP_1.0.9
  LIBDFP_1.0.10
  LIBDFP_1.0.13
  LIBDFP_1.0.15
  LIBDFP_1.0.16
  LIBDFP_PRIVATE
}
