/* Entry point for shared-libraries.

   Copyright (C) 2012-2015 Free Software Foundation, Inc.

   This file is part of the Decimal Floating Point C Library.

   The Decimal Floating Point C Library is free software; you can
   redistribute it and/or modify it under the terms of the GNU Lesser
   General Public License version 2.1.

   The Decimal Floating Point C Library is distributed in the hope that
   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
   the GNU Lesser General Public License version 2.1 for more details.

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 along with the Decimal Floating Point C Library;
   if not, write to the Free Software Foundation, Inc., 51 Franklin
   Street, Fifth Floor, Boston, MA 02110-1301 USA.

   Please see libdfp/COPYING.txt for more information.  */

#include <stdio.h>
#include <stdlib.h>
#include "config.h"

void __libdfp_version (void);

const char __invoke_dynamic_linker__[] __attribute__ ((section (".interp")))
 = LIBDFP_RUNTIME_LOADER;

void
__attribute__ ((noreturn))
__libdfp_version (void)
{
  printf ("Decimal Floating Point C Library (" PACKAGE_STRING").\n"
          "Copyright (C) 2014 Free Software Foundation.\n"
          "Compiled by GNU version "__VERSION__".\n"
          "For bug reporting please contact:\n"
          "<"PACKAGE_BUGREPORT">.\n");
  exit (0);
}

